@@ -42,7 +42,7 @@ APPLET_NAME = 'firewall-applet'
 DATADIR = '/usr/share/' + DAEMON_NAME
 CONFIG_GLADE_NAME = CONFIG_NAME + '.glade'
 COPYRIGHT = '(C) 2010-2017 Red Hat, Inc.'
-VERSION = '0.7.2'
+VERSION = '0.8.1'
 AUTHORS = [
     "Thomas Woerner <twoerner@redhat.com>",
     "Jiri Popelka <jpopelka@redhat.com>",
@@ -110,10 +110,8 @@ COMMANDS = {
     "eb":           "/usr/sbin/ebtables",
     "eb-restore":   "/usr/sbin/ebtables-restore",
     "ipset":        "/usr/sbin/ipset",
-    "modinfo":      "/usr/sbin/modinfo",
     "modprobe":     "/usr/sbin/modprobe",
     "rmmod":        "/usr/sbin/rmmod",
-    "nft":          "/usr/sbin/nft",
 }
 
 LOG_DENIED_VALUES = [ "all", "unicast", "broadcast", "multicast", "off" ]
@@ -128,7 +126,7 @@ FALLBACK_LOCKDOWN = False
 FALLBACK_IPV6_RPFILTER = True
 FALLBACK_INDIVIDUAL_CALLS = False
 FALLBACK_LOG_DENIED = "off"
-FALLBACK_AUTOMATIC_HELPERS = "system"
+FALLBACK_AUTOMATIC_HELPERS = "no"
 FALLBACK_FIREWALL_BACKEND = "nftables"
 FALLBACK_FLUSH_ALL_ON_RELOAD = True
 FALLBACK_RFC3964_IPV4 = True
