@@ -21,24 +21,28 @@
 
 __all__ = [ "Firewall" ]
 
-import os.path, sys
+import os.path
+import sys
 import copy
 import time
+import traceback
 from firewall import config
 from firewall import functions
 from firewall.core import ipXtables
 from firewall.core import ebtables
+from firewall.core import nftables
 from firewall.core import ipset
 from firewall.core import modules
 from firewall.core.fw_icmptype import FirewallIcmpType
 from firewall.core.fw_service import FirewallService
-from firewall.core.fw_zone import FirewallZoneIPTables
-from firewall.core.fw_direct import FirewallDirectIPTables
+from firewall.core.fw_zone import FirewallZone
+from firewall.core.fw_direct import FirewallDirect
 from firewall.core.fw_config import FirewallConfig
 from firewall.core.fw_policies import FirewallPolicies
 from firewall.core.fw_ipset import FirewallIPSet
 from firewall.core.fw_transaction import FirewallTransaction
 from firewall.core.fw_helper import FirewallHelper
+from firewall.core.fw_nm import nm_get_bus_name, nm_get_interfaces_in_zone
 from firewall.core.logger import log
 from firewall.core.io.firewalld_conf import firewalld_conf
 from firewall.core.io.direct import Direct
@@ -46,6 +50,7 @@ from firewall.core.io.service import service_reader
 from firewall.core.io.icmptype import icmptype_reader
 from firewall.core.io.zone import zone_reader, Zone
 from firewall.core.io.ipset import ipset_reader
+from firewall.core.ipset import IPSET_TYPES
 from firewall.core.io.helper import helper_reader
 from firewall import errors
 from firewall.errors import FirewallError
@@ -57,32 +62,38 @@ from firewall.errors import FirewallError
 ############################################################################
 
 class Firewall(object):
-    def __init__(self):
+    def __init__(self, offline=False):
         self._firewalld_conf = firewalld_conf(config.FIREWALLD_CONF)
+        self._offline = offline
 
-        self.ip4tables_backend = ipXtables.ip4tables()
-        self.ip4tables_enabled = True
-        self.ip4tables_supported_icmp_types = [ ]
-        self.ip6tables_backend = ipXtables.ip6tables()
-        self.ip6tables_enabled = True
-        self.ip6tables_supported_icmp_types = [ ]
-        self.ebtables_backend = ebtables.ebtables()
-        self.ebtables_enabled = True
-        self.ipset_backend = ipset.ipset()
-        self.ipset_enabled = True
-        self.ipset_supported_types = [ ]
+        if self._offline:
+            self.ip4tables_enabled = False
+            self.ip6tables_enabled = False
+            self.ebtables_enabled = False
+            self.ipset_enabled = False
+            self.ipset_supported_types = IPSET_TYPES
+            self.nftables_enabled = False
+        else:
+            self.ip4tables_backend = ipXtables.ip4tables(self)
+            self.ip4tables_enabled = True
+            self.ip4tables_supported_icmp_types = [ ]
+            self.ip6tables_backend = ipXtables.ip6tables(self)
+            self.ip6tables_enabled = True
+            self.ip6tables_supported_icmp_types = [ ]
+            self.ebtables_backend = ebtables.ebtables()
+            self.ebtables_enabled = True
+            self.ipset_backend = ipset.ipset()
+            self.ipset_enabled = True
+            self.ipset_supported_types = [ ]
+            self.nftables_backend = nftables.nftables(self)
+            self.nftables_enabled = True
 
-        self.available_tables = { }
-        self.available_tables["ipv4"] = self.ip4tables_backend.available_tables()
-        self.available_tables["ipv6"] = self.ip6tables_backend.available_tables()
-        self.available_tables["eb"] = self.ebtables_backend.available_tables()
-
-        self.modules_backend = modules.modules()
+            self.modules_backend = modules.modules()
 
         self.icmptype = FirewallIcmpType(self)
         self.service = FirewallService(self)
-        self.zone = FirewallZoneIPTables(self)
-        self.direct = FirewallDirectIPTables(self)
+        self.zone = FirewallZone(self)
+        self.direct = FirewallDirect(self)
         self.config = FirewallConfig(self)
         self.policies = FirewallPolicies()
         self.ipset = FirewallIPSet(self)
@@ -91,13 +102,12 @@ class Firewall(object):
         self.__init_vars()
 
     def __repr__(self):
-        return '%s(%r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r)' % \
+        return '%s(%r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r, %r)' % \
             (self.__class__, self.ip4tables_enabled, self.ip6tables_enabled,
              self.ebtables_enabled, self._state, self._panic,
              self._default_zone, self._module_refcount, self._marks,
-             self._min_mark, self.cleanup_on_exit, self.ipv6_rpfilter_enabled,
-             self.ipset_enabled, self._individual_calls, self._log_denied,
-             self._automatic_helpers)
+             self.cleanup_on_exit, self.ipv6_rpfilter_enabled,
+             self.ipset_enabled, self._individual_calls, self._log_denied)
 
     def __init_vars(self):
         self._state = "INIT"
@@ -106,15 +116,13 @@ class Firewall(object):
         self._module_refcount = { }
         self._marks = [ ]
         # fallback settings will be overloaded by firewalld.conf
-        self._min_mark = config.FALLBACK_MINIMAL_MARK
         self.cleanup_on_exit = config.FALLBACK_CLEANUP_ON_EXIT
         self.ipv6_rpfilter_enabled = config.FALLBACK_IPV6_RPFILTER
         self._individual_calls = config.FALLBACK_INDIVIDUAL_CALLS
         self._log_denied = config.FALLBACK_LOG_DENIED
-        self._automatic_helpers = config.FALLBACK_AUTOMATIC_HELPERS
-        self.nf_conntrack_helper_setting = 0
-        self.nf_conntrack_helpers = { }
-        self.nf_nat_helpers = { }
+        self._firewall_backend = config.FALLBACK_FIREWALL_BACKEND
+        self._flush_all_on_reload = config.FALLBACK_FLUSH_ALL_ON_RELOAD
+        self._rfc3964_ipv4 = config.FALLBACK_RFC3964_IPV4
 
     def individual_calls(self):
         return self._individual_calls
@@ -122,28 +130,29 @@ class Firewall(object):
     def _check_tables(self):
         # check if iptables, ip6tables and ebtables are usable, else disable
         if self.ip4tables_enabled and \
-           "filter" not in self.get_available_tables("ipv4"):
+           "filter" not in self.get_backend_by_ipv("ipv4").get_available_tables():
             log.warning("iptables not usable, disabling IPv4 firewall.")
             self.ip4tables_enabled = False
 
         if self.ip6tables_enabled and \
-           "filter" not in self.get_available_tables("ipv6"):
+           "filter" not in self.get_backend_by_ipv("ipv6").get_available_tables():
             log.warning("ip6tables not usable, disabling IPv6 firewall.")
             self.ip6tables_enabled = False
 
         if self.ebtables_enabled and \
-           "filter" not in self.get_available_tables("eb"):
+           "filter" not in self.get_backend_by_ipv("eb").get_available_tables():
             log.warning("ebtables not usable, disabling ethernet bridge firewall.")
             self.ebtables_enabled = False
 
         # is there at least support for ipv4 or ipv6
-        if not self.ip4tables_enabled and not self.ip6tables_enabled:
+        if not self.ip4tables_enabled and not self.ip6tables_enabled \
+           and not self.nftables_enabled:
             log.fatal("No IPv4 and IPv6 firewall.")
             sys.exit(1)
 
     def _start_check(self):
         try:
-            self.ipset_backend.list()
+            self.ipset_backend.set_list()
         except ValueError:
             log.warning("ipset not usable, disabling ipset usage in firewall.")
             # ipset is not usable, no supported types
@@ -151,7 +160,7 @@ class Firewall(object):
             self.ipset_supported_types = [ ]
         else:
             # ipset is usable, get all supported types
-            self.ipset_supported_types = self.ipset_backend.supported_types()
+            self.ipset_supported_types = self.ipset_backend.set_supported_types()
 
         self.ip4tables_backend.fill_exists()
         if not self.ip4tables_backend.restore_command_exists:
@@ -196,28 +205,6 @@ class Firewall(object):
             log.debug1("ebtables-restore is not supporting the --noflush "
                        "option, will therefore not be used")
 
-        if os.path.exists(config.COMMANDS["modinfo"]):
-            self.nf_conntrack_helpers = functions.get_nf_conntrack_helpers()
-            if len(self.nf_conntrack_helpers) > 0:
-                log.debug1("Conntrack helpers supported by the kernel:")
-                for key,values in self.nf_conntrack_helpers.items():
-                    log.debug1("  %s: %s", key, ", ".join(values))
-            else:
-                log.debug1("No conntrack helpers supported by the kernel.")
-
-            self.nf_nat_helpers = functions.get_nf_nat_helpers()
-            if len(self.nf_nat_helpers) > 0:
-                log.debug1("NAT helpers supported by the kernel:")
-                for key,values in self.nf_nat_helpers.items():
-                    log.debug1("  %s: %s", key, ", ".join(values))
-            else:
-                log.debug1("No NAT helpers supported by the kernel.")
-
-        else:
-            self.nf_conntrack_helpers = { }
-            self.nf_nat_helpers = { }
-            log.warning("modinfo command is missing, not able to detect conntrack helpers.")
-
     def _start(self, reload=False, complete_reload=False):
         # initialize firewall
         default_zone = config.FALLBACK_ZONE
@@ -233,13 +220,12 @@ class Firewall(object):
             if self._firewalld_conf.get("DefaultZone"):
                 default_zone = self._firewalld_conf.get("DefaultZone")
 
-            if self._firewalld_conf.get("MinimalMark"):
-                self._min_mark = int(self._firewalld_conf.get("MinimalMark"))
-
             if self._firewalld_conf.get("CleanupOnExit"):
                 value = self._firewalld_conf.get("CleanupOnExit")
                 if value is not None and value.lower() in [ "no", "false" ]:
                     self.cleanup_on_exit = False
+                log.debug1("CleanupOnExit is set to '%s'",
+                           self.cleanup_on_exit)
 
             if self._firewalld_conf.get("Lockdown"):
                 value = self._firewalld_conf.get("Lockdown")
@@ -277,21 +263,35 @@ class Firewall(object):
                     self._log_denied = value.lower()
                     log.debug1("LogDenied is set to '%s'", self._log_denied)
 
-            if self._firewalld_conf.get("AutomaticHelpers"):
-                value = self._firewalld_conf.get("AutomaticHelpers")
-                if value is not None:
-                    if value.lower() in [ "no", "false" ]:
-                        self._automatic_helpers = "no"
-                    elif value.lower() in [ "yes", "true" ]:
-                        self._automatic_helpers = "yes"
-                    else:
-                        self._automatic_helpers = value.lower()
-                    log.debug1("AutomaticHelpers is set to '%s'",
-                               self._automatic_helpers)
+            if self._firewalld_conf.get("FirewallBackend"):
+                self._firewall_backend = self._firewalld_conf.get("FirewallBackend")
+                log.debug1("FirewallBackend is set to '%s'",
+                           self._firewall_backend)
+
+            if self._firewalld_conf.get("FlushAllOnReload"):
+                value = self._firewalld_conf.get("FlushAllOnReload")
+                if value.lower() in [ "no", "false" ]:
+                    self._flush_all_on_reload = False
+                else:
+                    self._flush_all_on_reload = True
+                log.debug1("FlushAllOnReload is set to '%s'",
+                           self._flush_all_on_reload)
+
+            if self._firewalld_conf.get("RFC3964_IPv4"):
+                value = self._firewalld_conf.get("RFC3964_IPv4")
+                if value.lower() in [ "no", "false" ]:
+                    self._rfc3964_ipv4 = False
+                else:
+                    self._rfc3964_ipv4 = True
+                log.debug1("RFC3964_IPv4 is set to '%s'",
+                           self._rfc3964_ipv4)
 
         self.config.set_firewalld_conf(copy.deepcopy(self._firewalld_conf))
 
-        self._start_check()
+        self._select_firewall_backend(self._firewall_backend)
+
+        if not self._offline:
+            self._start_check()
 
         # load lockdown whitelist
         log.debug1("Loading lockdown whitelist")
@@ -375,11 +375,10 @@ class Firewall(object):
         self.direct.set_permanent_config(obj)
         self.config.set_direct(copy.deepcopy(obj))
 
-        # automatic helpers
-        if self._automatic_helpers != "system":
-            functions.set_nf_conntrack_helper_setting(self._automatic_helpers == "yes")
-        self.nf_conntrack_helper_setting = \
-            functions.get_nf_conntrack_helper_setting()
+        self._default_zone = self.check_zone(default_zone)
+
+        if self._offline:
+            return
 
         # check if needed tables are there
         self._check_tables()
@@ -410,6 +409,10 @@ class Firewall(object):
             log.debug1("Unloading firewall modules")
             self.modules_backend.unload_firewall_modules()
 
+        self.apply_default_tables(use_transaction=transaction)
+        transaction.execute(True)
+        transaction.clear()
+
         # apply settings for loaded ipsets while reloading here
         if self.ipset_enabled and self.ipset.has_ipsets():
             log.debug1("Applying ipsets")
@@ -425,7 +428,6 @@ class Firewall(object):
         log.debug1("Applying used zones")
         self.zone.apply_zones(use_transaction=transaction)
 
-        self._default_zone = self.check_zone(default_zone)
         self.zone.change_default_zone(None, self._default_zone,
                                       use_transaction=transaction)
 
@@ -594,7 +596,7 @@ class Firewall(object):
             except FirewallError as msg:
                 log.error("Failed to load %s file '%s': %s", reader_type,
                           name, msg)
-            except Exception as msg:
+            except Exception:
                 log.error("Failed to load %s file '%s':", reader_type, name)
                 log.exception()
 
@@ -611,11 +613,6 @@ class Firewall(object):
                 self.config.forget_zone(combined_zone.name)
             self.zone.add_zone(combined_zone)
 
-    def get_available_tables(self, ipv):
-        if ipv in [ "ipv4", "ipv6", "eb" ]:
-            return self.available_tables[ipv]
-        return [ ]
-
     def cleanup(self):
         self.icmptype.cleanup()
         self.service.cleanup()
@@ -629,29 +626,18 @@ class Firewall(object):
         self.__init_vars()
 
     def stop(self):
-        if self.cleanup_on_exit:
+        if self.cleanup_on_exit and not self._offline:
             self.flush()
             self.set_policy("ACCEPT")
             self.modules_backend.unload_firewall_modules()
 
         self.cleanup()
 
-    # marks
-
-    def new_mark(self):
-        # return first unused mark
-        i = self._min_mark
-        while i in self._marks:
-            i += 1
-        self._marks.append(i)
-        return i
-
-    def del_mark(self, mark):
-        self._marks.remove(mark)
-
     # handle modules
 
     def handle_modules(self, _modules, enable):
+        num_failed = 0
+        error_msgs = ""
         for i,module in enumerate(_modules):
             if enable:
                 (status, msg) = self.modules_backend.load_module(module)
@@ -661,9 +647,9 @@ class Firewall(object):
                 else:
                     (status, msg) = self.modules_backend.unload_module(module)
             if status != 0:
-                if enable:
-                    return (_modules[:i], msg) # cleanup modules and error msg
-                # else: ignore cleanup
+                num_failed += 1
+                error_msgs += msg
+                continue
 
             if enable:
                 self._module_refcount.setdefault(module, 0)
@@ -673,17 +659,101 @@ class Firewall(object):
                     self._module_refcount[module] -= 1
                     if self._module_refcount[module] == 0:
                         del self._module_refcount[module]
-        return None
+        return (num_failed, error_msgs)
 
-    # apply default rules
-    def __apply_default_rules(self, ipv, transaction):
-        if ipv is "ipv4":
-            x = self.ip4tables_backend
-        elif ipv is "ipv6":
-            x = self.ip6tables_backend
+    def _select_firewall_backend(self, backend):
+        if backend != "nftables":
+            self.nftables_enabled = False
+        # even if using nftables, the other backends are enabled for use with
+        # the direct interface. nftables is used for the firewalld primitives.
+
+    def get_backend_by_name(self, name):
+        for backend in self.all_backends():
+            if backend.name == name:
+                return backend
+        raise FirewallError(errors.UNKNOWN_ERROR,
+                            "'%s' backend does not exist" % name)
+
+    def get_backend_by_ipv(self, ipv):
+        if self.nftables_enabled:
+            return self.nftables_backend
+        if ipv == "ipv4" and self.ip4tables_enabled:
+            return self.ip4tables_backend
+        elif ipv == "ipv6" and self.ip6tables_enabled:
+            return self.ip6tables_backend
+        elif ipv == "eb" and self.ebtables_enabled:
+            return self.ebtables_backend
+        raise FirewallError(errors.INVALID_IPV,
+                            "'%s' is not a valid backend or is unavailable" % ipv)
+
+    def get_direct_backend_by_ipv(self, ipv):
+        if ipv == "ipv4" and self.ip4tables_enabled:
+            return self.ip4tables_backend
+        elif ipv == "ipv6" and self.ip6tables_enabled:
+            return self.ip6tables_backend
+        elif ipv == "eb" and self.ebtables_enabled:
+            return self.ebtables_backend
+        raise FirewallError(errors.INVALID_IPV,
+                            "'%s' is not a valid backend or is unavailable" % ipv)
+
+    def is_backend_enabled(self, name):
+        if name == "ip4tables":
+            return self.ip4tables_enabled
+        elif name == "ip6tables":
+            return self.ip6tables_enabled
+        elif name == "ebtables":
+            return self.ebtables_enabled
+        elif name == "nftables":
+            return self.nftables_enabled
+        return False
+
+    def is_ipv_enabled(self, ipv):
+        if self.nftables_enabled:
+            return True
+        if ipv == "ipv4":
+            return self.ip4tables_enabled
+        elif ipv == "ipv6":
+            return self.ip6tables_enabled
+        elif ipv == "eb":
+            return self.ebtables_enabled
+        return False
+
+    def enabled_backends(self):
+        backends = []
+        if self.nftables_enabled:
+            backends.append(self.nftables_backend)
         else:
-            x = self.ebtables_backend
-        x.apply_default_rules(transaction, self._log_denied)
+            if self.ip4tables_enabled:
+                backends.append(self.ip4tables_backend)
+            if self.ip6tables_enabled:
+                backends.append(self.ip6tables_backend)
+            if self.ebtables_enabled:
+                backends.append(self.ebtables_backend)
+        return backends
+
+    def all_backends(self):
+        backends = []
+        if self.ip4tables_enabled:
+            backends.append(self.ip4tables_backend)
+        if self.ip6tables_enabled:
+            backends.append(self.ip6tables_backend)
+        if self.ebtables_enabled:
+            backends.append(self.ebtables_backend)
+        if self.nftables_enabled:
+            backends.append(self.nftables_backend)
+        return backends
+
+    def apply_default_tables(self, use_transaction=None):
+        if use_transaction is None:
+            transaction = FirewallTransaction(self)
+        else:
+            transaction = use_transaction
+
+        for backend in self.enabled_backends():
+            transaction.add_rules(backend, backend.build_default_tables())
+
+        if use_transaction is None:
+            transaction.execute(True)
 
     def apply_default_rules(self, use_transaction=None):
         if use_transaction is None:
@@ -691,30 +761,23 @@ class Firewall(object):
         else:
             transaction = use_transaction
 
-        for ipv in [ "ipv4", "ipv6", "eb" ]:
-            self.__apply_default_rules(ipv, transaction)
+        for backend in self.enabled_backends():
+            rules = backend.build_default_rules(self._log_denied)
+            transaction.add_rules(backend, rules)
 
-        if self.ipv6_rpfilter_enabled and \
-           "raw" in self.get_available_tables("ipv6"):
+        if self.is_ipv_enabled("ipv6"):
+            ipv6_backend = self.get_backend_by_ipv("ipv6")
+            if "raw" in ipv6_backend.get_available_tables():
+                if self.ipv6_rpfilter_enabled:
+                    rules = ipv6_backend.build_rpfilter_rules(self._log_denied)
+                    transaction.add_rules(ipv6_backend, rules)
 
-            # Execute existing transaction
+        if self.is_ipv_enabled("ipv6") and self._rfc3964_ipv4:
+            rules = ipv6_backend.build_rfc3964_ipv4_rules()
+            transaction.add_rules(ipv6_backend, rules)
+
+        if use_transaction is None:
             transaction.execute(True)
-            # Start new transaction
-            transaction.clear()
-
-            self.ip6tables_backend.apply_rpfilter_rules(transaction, self._log_denied)
-
-            # Execute ipv6_rpfilter transaction, it might fail
-            try:
-                transaction.execute(True)
-            except FirewallError as msg:
-                log.warning("Applying rules for ipv6_rpfilter failed: %s", msg)
-            # Start new transaction
-            transaction.clear()
-
-        else:
-            if use_transaction is None:
-                transaction.execute(True)
 
     # flush and policy
 
@@ -726,26 +789,14 @@ class Firewall(object):
 
         log.debug1("Flushing rule set")
 
-        if self.ip4tables_enabled:
-            try:
-                self.ip4tables_backend.flush(transaction)
-            except Exception as e:
-                log.error("Failed to flush ipv4 firewall: %s" % e)
-        if self.ip6tables_enabled:
-            try:
-                self.ip6tables_backend.flush(transaction)
-            except Exception as e:
-                log.error("Failed to flush ipv6 firewall: %s" % e)
-        if self.ebtables_enabled:
-            try:
-                self.ebtables_backend.flush(transaction)
-            except Exception as e:
-                log.error("Failed to flush eb firewall: %s" % e)
+        for backend in self.all_backends():
+            rules = backend.build_flush_rules()
+            transaction.add_rules(backend, rules)
 
         if use_transaction is None:
             transaction.execute(True)
 
-    def set_policy(self, policy, which="used", use_transaction=None):
+    def set_policy(self, policy, use_transaction=None):
         if use_transaction is None:
             transaction = FirewallTransaction(self)
         else:
@@ -753,105 +804,58 @@ class Firewall(object):
 
         log.debug1("Setting policy to '%s'", policy)
 
-        if self.ip4tables_enabled:
-            try:
-                self.ip4tables_backend.set_policy(policy, which, transaction)
-            except Exception as e:
-                log.error("Failed to set policy of ipv4 firewall: %s" % e)
-
-        if self.ip6tables_enabled:
-            try:
-                self.ip6tables_backend.set_policy(policy, which, transaction)
-            except Exception as e:
-                log.error("Failed to set policy of ipv6 firewall: %s" % e)
-        if self.ebtables_enabled:
-            try:
-                self.ebtables_backend.set_policy(policy, which, transaction)
-            except Exception as e:
-                log.error("Failed to set policy of eb firewall: %s" % e)
+        for backend in self.enabled_backends():
+            rules = backend.build_set_policy_rules(policy)
+            transaction.add_rules(backend, rules)
 
         if use_transaction is None:
             transaction.execute(True)
 
     # rule function used in handle_ functions
 
-    def rule(self, ipv, rule):
-        # remove leading and trailing '"' for use with execve
-        i = 0
-        while i < len(rule):
-            x = rule[i]
-            if len(x) > 2 and x[0] == '"' and x[-1] == '"':
-                rule[i] = x[1:-1]
-            i += 1
-
-        if ipv == "ipv4":
-            # do not call if disabled
-            if self.ip4tables_enabled:
-                return self.ip4tables_backend.set_rule(rule, self._log_denied)
-        elif ipv == "ipv6":
-            # do not call if disabled
-            if self.ip6tables_enabled:
-                return self.ip6tables_backend.set_rule(rule, self._log_denied)
-        elif ipv == "eb":
-            # do not call if disabled
-            if self.ebtables_enabled:
-                return self.ebtables_backend.set_rule(rule, self._log_denied)
-        else:
-            raise FirewallError(errors.INVALID_IPV,
-                                "'%s' not in {'ipv4'|'ipv6'|'eb'}" % ipv)
-
-        return ""
-
-    def rules(self, ipv, rules):
-        _rules = rules[:]
-
-        backend = None
-        if ipv == "ipv4":
-            if self.ip4tables_enabled:
-                # do not call if disabled
-                backend = self.ip4tables_backend
-        elif ipv == "ipv6":
-            if self.ip6tables_enabled:
-                # do not call if disabled
-                backend = self.ip6tables_backend
-        elif ipv == "eb":
-            if self.ebtables_enabled:
-                # do not call if disabled
-                backend = self.ebtables_backend
-        else:
-            raise FirewallError(errors.INVALID_IPV,
-                                "'%s' not in {'ipv4'|'ipv6'|'eb'}" % ipv)
-
-        if not backend:
+    def rule(self, backend_name, rule):
+        if not rule:
             return ""
 
+        backend = self.get_backend_by_name(backend_name)
+        if not backend:
+            raise FirewallError(errors.INVALID_IPV,
+                                "'%s' is not a valid backend" % backend_name)
+
+        if not self.is_backend_enabled(backend_name):
+            return ""
+
+        return backend.set_rule(rule, self._log_denied)
+
+    def rules(self, backend_name, rules):
+        _rules = list(filter(None, rules))
+
+        backend = self.get_backend_by_name(backend_name)
+        if not backend:
+            raise FirewallError(errors.INVALID_IPV,
+                                "'%s' is not a valid backend" % backend_name)
+
+        if not self.is_backend_enabled(backend_name):
+            return
+
         if self._individual_calls or \
            not backend.restore_command_exists or \
-           (ipv == "eb" and not self.ebtables_backend.restore_noflush_option):
+           (backend_name == "ebtables" and not self.ebtables_backend.restore_noflush_option):
             for i,rule in enumerate(_rules):
-                # remove leading and trailing '"' for use with execve
-                j = 0
-                while j < len(rule):
-                    x = rule[j]
-                    if len(x) > 2 and x[0] == '"' and x[-1] == '"':
-                        rule[j] = x[1:-1]
-                    j += 1
-
                 try:
-                    backend.set_rule(rule)
+                    backend.set_rule(rule, self._log_denied)
                 except Exception as msg:
-                    log.error("Failed to apply rules. A firewall reload might solve the issue if the firewall has been modified using ip*tables or ebtables.")
+                    log.debug1(traceback.format_exc())
                     log.error(msg)
                     for rule in reversed(_rules[:i]):
                         try:
-                            backend.set_rule(ipXtables.reverse_rule(rule))
+                            backend.set_rule(backend.reverse_rule(rule), self._log_denied)
                         except Exception:
                             # ignore errors here
                             pass
-                    return False
-            return True
+                    raise msg
         else:
-            return backend.set_rules(_rules, log_denied=self._log_denied)
+            backend.set_rules(_rules, self._log_denied)
 
     # check functions
 
@@ -916,19 +920,24 @@ class Firewall(object):
     def reload(self, stop=False):
         _panic = self._panic
 
-        # save zone interfaces
-        _zone_interfaces = { }
-        for zone in self.zone.get_zones():
-            _zone_interfaces[zone] = self.zone.get_settings(zone)["interfaces"]
-        # save direct config
-        _direct_config = self.direct.get_runtime_config()
-        _old_dz = self.get_default_zone()
+        # must stash this. The value may change after _start()
+        flush_all = self._flush_all_on_reload
+
+        if not flush_all:
+            # save zone interfaces
+            _zone_interfaces = { }
+            for zone in self.zone.get_zones():
+                _zone_interfaces[zone] = self.zone.get_settings(zone)["interfaces"]
+            # save direct config
+            _direct_config = self.direct.get_runtime_config()
+            _old_dz = self.get_default_zone()
+
+        if not _panic:
+            self.set_policy("DROP")
 
         # stop
         self.cleanup()
 
-        self.set_policy("DROP")
-
         start_exception = None
         try:
             self._start(reload=True, complete_reload=stop)
@@ -937,47 +946,52 @@ class Firewall(object):
             # etc. We'll re-raise it at the end.
             start_exception = e
 
-        # handle interfaces in the default zone and move them to the new
-        # default zone if it changed
-        _new_dz = self.get_default_zone()
-        if _new_dz != _old_dz:
-            # if_new_dz has been introduced with the reload, we need to add it
-            # https://github.com/firewalld/firewalld/issues/53
-            if _new_dz not in _zone_interfaces:
-                _zone_interfaces[_new_dz] = { }
-            # default zone changed. Move interfaces from old default zone to
-            # the new one.
-            for iface, settings in list(_zone_interfaces[_old_dz].items()):
-                if settings["__default__"]:
-                    # move only those that were added to default zone
-                    # (not those that were added to specific zone same as
-                    # default)
-                    _zone_interfaces[_new_dz][iface] = \
-                        _zone_interfaces[_old_dz][iface]
-                    del _zone_interfaces[_old_dz][iface]
+        if not flush_all:
+            # handle interfaces in the default zone and move them to the new
+            # default zone if it changed
+            _new_dz = self.get_default_zone()
+            if _new_dz != _old_dz:
+                # if_new_dz has been introduced with the reload, we need to add it
+                # https://github.com/firewalld/firewalld/issues/53
+                if _new_dz not in _zone_interfaces:
+                    _zone_interfaces[_new_dz] = { }
+                # default zone changed. Move interfaces from old default zone to
+                # the new one.
+                for iface, settings in list(_zone_interfaces[_old_dz].items()):
+                    if settings["__default__"]:
+                        # move only those that were added to default zone
+                        # (not those that were added to specific zone same as
+                        # default)
+                        _zone_interfaces[_new_dz][iface] = \
+                            _zone_interfaces[_old_dz][iface]
+                        del _zone_interfaces[_old_dz][iface]
 
-        # add interfaces to zones again
-        for zone in self.zone.get_zones():
-            if zone in _zone_interfaces:
-                self.zone.set_settings(zone, { "interfaces":
-                                               _zone_interfaces[zone] })
-                del _zone_interfaces[zone]
-            else:
-                log.info1("New zone '%s'.", zone)
-        if len(_zone_interfaces) > 0:
-            for zone in list(_zone_interfaces.keys()):
-                log.info1("Lost zone '%s', zone interfaces dropped.", zone)
-                del _zone_interfaces[zone]
-        del _zone_interfaces
+            # add interfaces to zones again
+            for zone in self.zone.get_zones():
+                if zone in _zone_interfaces:
+                    self.zone.set_settings(zone, { "interfaces":
+                                                   _zone_interfaces[zone] })
+                    del _zone_interfaces[zone]
+                else:
+                    log.info1("New zone '%s'.", zone)
+            if len(_zone_interfaces) > 0:
+                for zone in list(_zone_interfaces.keys()):
+                    log.info1("Lost zone '%s', zone interfaces dropped.", zone)
+                    del _zone_interfaces[zone]
+            del _zone_interfaces
 
-        # restore direct config
-        self.direct.set_config(_direct_config)
+            # restore direct config
+            self.direct.set_config(_direct_config)
 
-        # enable panic mode again if it has been enabled before or set policy
-        # to ACCEPT
-        if _panic:
-            self.enable_panic_mode()
-        else:
+        # Restore permanent interfaces from NetworkManager
+        nm_bus_name = nm_get_bus_name()
+        if nm_bus_name:
+            for zone in self.zone.get_zones() + [""]:
+                for interface in nm_get_interfaces_in_zone(zone):
+                    self.zone.change_zone_of_interface(zone, interface, sender=nm_bus_name)
+
+        self._panic = _panic
+        if not self._panic:
             self.set_policy("ACCEPT")
 
         if start_exception:
@@ -998,9 +1012,8 @@ class Firewall(object):
             raise FirewallError(errors.ALREADY_ENABLED,
                                 "panic mode already enabled")
 
-        # TODO: use rule in raw table not default chain policy
         try:
-            self.set_policy("DROP", "all")
+            self.set_policy("PANIC")
         except Exception as msg:
             raise FirewallError(errors.COMMAND_FAILED, msg)
         self._panic = True
@@ -1010,9 +1023,8 @@ class Firewall(object):
             raise FirewallError(errors.NOT_ENABLED,
                                 "panic mode is not enabled")
 
-        # TODO: use rule in raw table not default chain policy
         try:
-            self.set_policy("ACCEPT", "all")
+            self.set_policy("ACCEPT")
         except Exception as msg:
             raise FirewallError(errors.COMMAND_FAILED, msg)
         self._panic = False
@@ -1038,24 +1050,6 @@ class Firewall(object):
         else:
             raise FirewallError(errors.ALREADY_SET, value)
 
-    # AUTOMATIC HELPERS
-
-    def get_automatic_helpers(self):
-        return self._automatic_helpers
-
-    def set_automatic_helpers(self, value):
-        if value not in config.AUTOMATIC_HELPERS_VALUES:
-            raise FirewallError(errors.INVALID_VALUE,
-                                "'%s', choose from '%s'" % \
-                                (value, "','".join(config.AUTOMATIC_HELPERS_VALUES)))
-
-        if value != self.get_automatic_helpers():
-            self._automatic_helpers = value
-            self._firewalld_conf.set("AutomaticHelpers", value)
-            self._firewalld_conf.write()
-        else:
-            raise FirewallError(errors.ALREADY_SET, value)
-
     # DEFAULT ZONE
 
     def get_default_zone(self):
